from operation import operation_factory
from datetime import datetime


class Revision(object):
    """
    Revision class
    Corresponds to Revision resource type in Google API: 
    https://developers.google.com/drive/v3/reference/revisions
    A Revision contains an Operation
    """

    def __init__(self, revision_raw):
        # timestamp
        self.time = datetime.fromtimestamp(revision_raw[1] / 1e3)
        # google user id of the revision author
        self.user_id = revision_raw[2]
        # numeric revision id, corresponding to id in the Google API
        self.revision_id = revision_raw[3]
        # session string identifier
        self.session_id = revision_raw[4]
        # index of the revision, within the user session
        self.session_revision_index = revision_raw[5]
        # array of raw revision metadata
        self.raw = revision_raw
        # dictionary of raw operation metadata
        self.operation_raw = revision_raw[0]
        # Operation object
        self.operation = operation_factory(self.operation_raw, self)
        # Iterator for operations
        self.iter_operations = self.operation.iter_operations
        # Iterator for suboperations
        self.iter_suboperations = self.operation.iter_suboperations
        
    @property
    def operations(self):
        """
        List of operations that make up this revision
        (MultiOperations are flattened into their base operations)
        """
        return list(self.iter_operations())

    @property
    def suboperations(self):
        """
        List of suboperations that make up this revision
        """
        return list(self.iter_suboperations())

    def apply(self, elements):
        """
        Apply the revision to a list of elements

        Arguments:
            elements (list): usually the elements attribute of a Content instance
        """
        for operation in self.iter_operations():
            operation.apply(elements)
        
    def to_dict(self):
        DICT_ATTRIBUTES = [
            'time',
            'user_id',
            'revision_id',
            'session_id',
            'session_revision_index',
            'operation',
        ]
        return {attr:getattr(self,attr) for attr in DICT_ATTRIBUTES}



from apiclient.discovery import build
from httplib2 import Http
import json
from revision import Revision
from session import Session
import pickle
from collections import defaultdict
import copy
import logging
from element import EndOfBody
from oauth2client.client import OAuth2Credentials
from oauth2client.service_account import ServiceAccountCredentials

logger = logging.getLogger('gdocrevisions')

# suppress warnings from google api client library
logging.getLogger('googleapiclient.discovery_cache').setLevel(logging.ERROR)

class Content(object):
    """
    Represents document content with a list of Element objects
    """
    def __init__(self):
        self.elements = [EndOfBody()]

    def apply(self, change):
        """
        Apply some change (could be a revision, operation, or suboperation) to the content elements

        Arguments:
            change (Revision, Operation, or Suboperation): object whoses changes should be applied to the content instance
        """
        change.apply(self.elements)
        
    def render(self):
        return ''.join([element.render() for element in self.elements])

    def reset(self):
        self.__init__()


class Document(object):
    """
    Basic Document class

    Attributes:
        revisions
        content
        current_revision_id
        operations
        sessions
    """
    def __init__(self, revisions, apply_revisions=True):
        """
        Arguments:
            revisions (list): list of revision objects
            latest (bool): indicates whether to build the content state on initialization
        """
        logger.debug("[Document.__init__()]: Start")
        self.revisions = revisions
        """ List of Revision objects """
        self.content = Content()
        """ Content object """
        self.current_revision_id = self.revisions[-1].revision_id if len(self.revisions)>1 else 1
        # populate content by applying all revisions
        if apply_revisions:
            logger.debug("[Document.__init__()]: Applying revisions")
            for revision in self.revisions:
                self.content.apply(revision)
        logger.debug("[Document.__init__()]: Finished")

    def at_time(self, datetime):
        """
        Revert document content to a point in time 
        """
        revisions = filter(lambda revision: revision.time<=datetime, self.revisions)
        self.content.reset()
        self.content.apply(revisions)
        return self

    def at_revision(self, revision_id):
        """
        Revert document content to a revision id
        """
        revisions = filter(lambda revision: revision.revision_id<=revision_id, self.revisions)
        self.content.reset()
        self.content.apply(revisions)
        return self

    def to_pickle(self, path):
        """
        Pickle a Document
        """
        with open(path, 'wb') as f:
            document = pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)

    @property
    def sessions(self):
        """
        Return a list of Session objects
        """
        sessions = []
        session_revisions = defaultdict(list)
        for revision in self.revisions:
            session_revisions[revision.session_id].append(revision)
        for session_id, revisions in session_revisions.items():
            if session_id:
                sessions.append(Session(revisions))
        # sort by start time attribute
        sessions.sort(key=lambda x: x.start_time)
        return sessions

    def replay(self, by="revision"):
        """
        iterator/generator that returns document states
        """
        content_state = Content()
        if by == 'operation':
            for revision in self.revisions:
                content_state.apply(revision)
                yield content_state
        elif by == 'revision':
            for operation in self.iter_operations():
                content_state.apply(operation)
                yield content_state
        else:
            raise ValueError("'by' not in accepted values (revision, operation)")

    def iter_operations(self):
        """
        Generator that iterates over operations
        """
        for revision in self.revisions:
            for operation in revision.iter_operations():
                yield operation

    def iter_suboperations(self):
        """
        Generator that iterates over suboperations
        """
        for revision in self.revisions:
            for suboperation in revision.iter_suboperations():
                yield suboperation

    @property
    def operations(self):
        """
        List of all operations that make up this document
        (MultiOperations are flattened into their base operations)
        """
        return list(self.iter_operations())

    @property
    def suboperations(self):
        """
        List of all suboperations that make up this document
        """
        return list(self.iter_suboperations())


class GoogleDoc(Document):
    """
    Google doc class
    Contains document metadata and revision history
    """
    def __init__(self, file_id, credentials=None, keyfile=None, **kwargs):
        """
        Create a GoogleDoc instance
        Requires either credentials or keyfile arguments to be specified

        Arguments:
            file_id (str): ID string that can be found in the Google Doc URL
            credentials (oauth2client.OAuth2Credentals): credentials object
            keyfile (str): Path to a service account json keyfile
        """
        logger.debug("[GoogleDoc.__init__()]: Start")
        # google credentials object instance (oauth2client.OAuth2Credentials or subclass)
        self.credentials = self._get_credentials(credentials, keyfile)
        # dictionary of document metadata via Google API
        self.metadata = self._gdrive_api().files().get(fileId=file_id).execute()
        # doc title
        self.name = self.metadata['name']
        # file identifier string from the URL
        self.file_id = file_id
        # dict of raw revision metadata, containing keys "changelog" and "chunkedSnapshot"
        self.revisions_raw = self._download_revision_details()
        # array of Revision objects
        logger.debug("[GoogleDoc.__init__()]: Initializing revisions")
        revisions = [Revision(r) for r in self.revisions_raw['changelog']]
        # initialize Document attributes
        super(GoogleDoc, self).__init__(revisions, **kwargs)

    def _get_credentials(self, credentials, keyfile):
        if credentials:
            if isinstance(credentials, OAuth2Credentials):
                return credentials
            else:
                raise TypeError("Credential object is not a valid OAuth2Credentials object")
        elif keyfile:
            scope = ['https://www.googleapis.com/auth/drive']
            return ServiceAccountCredentials.from_json_keyfile_name(keyfile, scope)
        else:
            raise ValueError("No credentials provided")

    def _gdrive_api(self):
        """
        Return an authorized drive api service object
        """
        return build('drive', 'v3', credentials=self.credentials)
        
    def _last_revision_id(self):
        """
        Return the id of the last revision to a document, using the offical google api v3
        """
        revision_metainfo = self._gdrive_api().revisions().list(fileId=self.file_id).execute()
        if len(revision_metainfo['revisions']) == 1:
            return revision_metainfo['revisions'][0]['id']
        else:
            return revision_metainfo['revisions'][-1]['id']
    
    def _generate_revision_url(self, start, end):
        """
        Generates a url for downloading revision details (using undocumented google api endpoint)
        """
        base_url = 'https://docs.google.com/document/d/{file_id}/revisions/load?id={file_id}&start={start}&end={end}'
        url = base_url.format(file_id=self.file_id,start=start,end=end)
        return url

    def _download_revision_details(self):
        """
        download json-like data with revision info
        """
        http_auth = self.credentials.authorize(Http())
        last_revision_id = self._last_revision_id()
        url = self._generate_revision_url(start=1,end=last_revision_id)
        raw_text = http_auth.request(url)[1][5:]
        return json.loads(raw_text)


def read_pickle(path):
    """
    Load Document from pickle
    """
    with open(path, 'rb') as f:
        document = pickle.load(f)
    return document




class Element(object):
	"""
	Base Element class. 
	Is a list element in a document's content attribute.
	Associated with a revision
	"""
	def __init__(self, revision):
		self.revision = revision

	def render(self):
		return ""


class Character(Element):
	"""
	Character element
	Represents a single character
	"""
	def __init__(self, char, revision):
		# revision object reference
		self.revision = revision
		# character string
		self.char = char

	def render(self):
		return self.char


class EndOfBody(Element):
	"""
	Dummy element indicating end of document body
	Separates main content from footnote text
	"""
	def __init__(self):
		pass



from element import Character

class Suboperation(object):
    """
    Base Suboperation class
    Represents action that are associated with single Element
    """
    def __init__(self, element):
        """
        operation_raw is a dictionary of raw operation metadata
        """
        self.element = element

    @property
    def type(self):
        return self.__class__.__name__

    def apply(self, elements):
        """
        elements is a list of Element instances
        """
        pass


class InsertElement(Suboperation):
    """
    Insert Character Suboperation class
    """
    def __init__(self, index, element):
        self.index = index
        self.element = element

    def apply(self, elements):
        elements.insert(self.index-1,self.element)


class DeleteElement(Suboperation):
    """
    Delete Character Suboperation class
    """
    def __init__(self, index):
        self.index = index
        # self.element = element

    def apply(self, elements):
        element = elements.pop(self.index-1)

        # TODO could modify popped element here?
        # would revision need to be passed in then?



from setuptools import setup

setup(name='gdocrevisions',
      version='0.4',
      description='Package for downloading and analyzing google doc revision history data.',
      url='https://github.com/harvard-vpal/gdocrevisions',
      author='Andrew Ang',
      author_email='andrew_ang@harvard.edu',
      license='MIT',
      packages=['gdocrevisions'],
      install_requires=[
      	'oauth2client',
      	'google-api-python-client',
      ],
)



from element import Character
from suboperation import InsertElement, DeleteElement


def operation_factory(operation_raw, revision):
    """
    Factory method that returns Operation or subclass instance

    Arguments:
        operation_raw (dict): raw operation data
        revision (Revision): revision instance, which gets associated with any created Element objects
    """
    operation_types = {
        'is': InsertString,
        'ds': DeleteString,
        'mlti': MultiOperation,
    }
    operation = operation_types.get(operation_raw['ty']) or Operation
    return operation(operation_raw, revision)


class Operation(object):
    """
    Base Operation class
    Represents action(s) that occur as part of a revision
    """
    def __init__(self, operation_raw, revision):
        """
        operation_raw is a dictionary of raw operation metadata
        """
        # could remove self.raw (dict) to save memory
        self.raw = operation_raw
        self.type = self.__class__.__name__
        self.revision = revision
        self.suboperations = []

    def apply(self, elements):
        """
        elements is a list of Element instances
        """
        for suboperation in self.suboperations:
            suboperation.apply(elements)

    def iter_operations(self):
        yield self

    def iter_suboperations(self):
        for suboperation in self.suboperations:
            yield suboperation

    def to_dict(self):
        DICT_ATTRIBUTES = [
            'type',
            'revision',
        ]
        return {attr:getattr(self,attr) for attr in DICT_ATTRIBUTES}


class InsertString(Operation):
    """
    Operation subclass representing an "Insert String" operation
    """
    def __init__(self, operation_raw, revision):
        super(InsertString, self).__init__(operation_raw, revision)
        self.string = operation_raw['s']
        self.start_index = operation_raw['ibi']
        # self.elements = [Character(self.revision, char) for char in self.string]
        self.suboperations = [InsertElement(self.start_index+i, Character(char, self.revision)) for i,char in enumerate(self.string)]


class DeleteString(Operation):
    """
    Operation subclass representing a "Delete String" operation
    """
    def __init__(self, operation_raw, revision):
        super(DeleteString, self).__init__(operation_raw, revision)
        self.start_index = operation_raw['si']
        self.end_index = operation_raw['ei']
        self.suboperations = [DeleteElement(self.end_index-i) for i in range(self.end_index-self.start_index+1)]


class MultiOperation(Operation):
    """
    Operation subclass representing a "Multiple Operation" operation
    Contains an array of Operation objects
    """
    def __init__(self, operation_raw, revision):
        super(MultiOperation, self).__init__(operation_raw, revision)
        self.operations = [operation_factory(operation_raw, revision) for operation_raw in operation_raw['mts']]

    def apply(self, elements):
        """
        Apply each of the operations comprising the MultiOperation
        """
        for operation in self.operations:
            self.operation.apply(elements)

    def iter_operations(self):
        """
        Generator that iterates through base operations.
        Does a depth-first search of the operations tree,
        yielding leaf (non-Multioperation) nodes
        """
        for operation in self.operations:
            if hasattr(operation, 'operations'):
                for operation in operation.iter_operations():
                    yield operation
            else:
                yield operation

    def iter_suboperations(self):
        for operation in self.iter_operations():
            for suboperation in operation.iter_suboperations():
                yield suboperation



class Session(object):
	"""
	Session class
	"""
	def __init__(self, revisions):
		# list of revisions
		self.revisions = revisions
		# session id
		self.session_id = self.revisions[0]
		# time of first revision
		self.start_time = self.revisions[0].time
		# time of last revision
		self.end_time = self.revisions[-1].time if len(self.revisions)>1 else self.start_time
		# user
		self.user_id  = self.revisions[0].user_id


